#!/usr/bin/env python
import sys
sys.path.append('~/.local/lib/python3.2')
import json
import pprint as pp
from collections import UserList
import copy as cp
import time

from mpi4py import rc
#rc.threaded = False
#rc.profile('vt-mpi', logfile='cpilog')
#rc.profile('mpe', logfile='cpilog')
from mpi4py import MPI

combinations_tried = 0
business_length = 4
economy_length = 6
maximum_opt = 0

class Passenger:
    def __init__(self, passenger_hash):
        self.id = passenger_hash['ID']
        self.group_id = passenger_hash["group_id"]
        self.seat_class = passenger_hash['class']
        self.destination = passenger_hash['destination']
        self.seat_pref = passenger_hash['seat_pref']
        self.time = passenger_hash['time']

    def __str__(self):
        return "[{id}] {sp} seat in {sc} class, {fltime} to {fldest}".format(
            id=self.id,
            sp=self.seat_pref,
            sc=self.seat_class,
            fltime=self.time,
            fldest=self.destination)

    def __repr__(self):
        return "Passenger: "+str(self.id)+" to "+str(self.destination)+" at "+str(self.time)

    def __eq__(self, other):
        if type(self) is type(other):
            return self.id == other.id

    def __ne__(self, other):
        if type(self) is type(other):
            return not self.id == other.id

    def size(self):
        return 1


class PassengerGroup:
    def __init__(self, passengerlist):
        self.passengers = passengerlist
        self.groupsize = len(passengerlist)
        self.destination = passengerlist[0].destination
        self.time = passengerlist[0].time
        self.seat_class = passengerlist[0].seat_class

    def add_passenger(self, passenger):
        if isinstance(passenger, Passenger):
            self.passengers.append(passenger)
        else:
            raise TypeError

    def __add__(self, other):
        if isinstance(other, Passenger):
            self.passengers.append(other)
        else:
            raise TypeError

    def __str__(self):
        result = "{num} passengers, {fltime} to {fldest}\n".format(
            num=self.groupsize,
            fltime=self.time,
            fldest=self.destination)
        for passenger in self.passengers:
            result += "\t{}\n".format(str(passenger))
        return result

    def __repr__(self):
        return str(self)

    def size(self):
        return self.groupsize


class Flight:
    def __init__(self, flight_hash):
        self.id = flight_hash['id']
        self.destination = flight_hash['destination']
        self.time = flight_hash['time']
        # self.seats = {
        #     'total': flight_hash['seats_total'],
        #     'first': flight_hash['seats_classes']['first'],
        #     'business': flight_hash['seats_classes']['business'],
        #     'economy': flight_hash['seats_classes']['economy']
        # }
        self.seatmap = Seatmap(flight_hash['planetype'])

    def __str__(self):
        return "Flight {id} to {destination} at {time}".format(
            id=self.id,
            destination=self.destination,
            time=self.time
        )

    def __repr__(self):
        return str(self)

class SeatRow(UserList):
    def __init__(self):
        self.data = []

    def free_seats_count():
        count = 0
        for seat in self.data:
            if(seat.occupied):
                ++count
        return count


class Seatmap(UserList):
    def __init__(self, planetype):
        global business_length
        global economy_length
        self.data = []
        self.planetype = planetype
        self.placed = 0
        if planetype == "LH-A320":
            self.business_seats = ['A', 'C', 'D', 'F']
            self.economy_seats = ['A', 'B', 'C', 'D', 'E', 'F']
            #self.business_range = range(2)
            #self.economy_range = range(2,4)
            self.business_range = range(business_length)
            self.economy_range = range(business_length,economy_length)
            # 30 rows; first 7 business; 23 economy
            for rownum in self.business_range:
                row = SeatRow()
                for i in range(4):
                    letter = self.business_seats[i]
                    if(letter == "C" or letter == "D"):
                        seattype = "aisle"
                    elif(letter == "A" or letter == "F"):
                        seattype = "window"
                    seat = Seat(str(rownum) + letter, letter, (rownum, i), "business")
                    row.append(seat)
                self.data.append(row)
            for rownum in self.economy_range:
                row = SeatRow()
                for i in range(6):
                    letter = self.economy_seats[i]
                    if(letter == "C" or letter == "D"):
                        seattype = "aisle"
                    elif(letter == "A" or letter == "F"):
                        seattype = "window"
                    elif(letter == "B" or letter == "E"):
                        seattype = "middle"
                    seat = Seat(str(rownum) + letter, letter, (rownum, i), "economy")
                    row.append(seat)
                self.data.append(row)

    def __repr__(self):
        string = "\n"
        for row in self.data:
            string+="|"
            for seat in row:
                if(seat.occupied):
                    string+=str(seat.passenger.id)+" "
                else:
                    string+="-"
            string+="\n"
        for i in range(10):
            string+="_"
        return string


    def occupy_by_xy(self,passenger, row_nr, seat_nr):
        self.data[row_nr][seat_nr].occupy(passenger)
        self.placed+=1

    def get_coords_for_id(self, seatid):
        nr = ""
        for char in seatid:
            if(is_number(char)):
                nr+=char
            else:
                seat_letter = char
        row_nr = int(nr)
        row = self.data[row_nr]
        if(row_nr in self.business_range):
            seat_letter = seatid[-1]
            seat_nr = self.business_seats.index(seat_letter)
        elif(row_nr in self.economy_range):
            seat_letter = seatid[-1]
            seat_nr = self.economy_seats.index(seat_letter)
        return row_nr, seat_nr

    def occupy_by_id(self, passenger, seatid):
        row_nr, seat_nr = self.get_coords_for_id(seatid)
        self.occupy_by_xy(passenger,row_nr, seat_nr)

    def next_free_seat(self):
        stop = False
        for row in self.data:
            for seat in row:
                if(not seat.occupied):
                    stop = True
                    return seat
                    break
            if(stop):
                break
        return False


class Seat:
    def __init__(self, seatid, seat_letter, seat_coord, seatclass="economy"):
        self.id = seatid
        self.occupied = False
        self.seatclass = seatclass
        self.seat_letter = seat_letter
        letter_map = {"A": "window", "F": "window", "B": "middle", "E": "middle", "C": "aisle", "D": "aisle"}
        self.seat_pref = letter_map[self.seat_letter]
        self.coord = seat_coord

    def occupy(self,passenger):
        self.occupied = True
        self.passenger = passenger

    def get_neigbours(self, seatmap):
        global economy_length
        global business_length
        neighbours = []
        x, y = self.coord
        if(x>0 and not x==business_length):
            neighbours.append(seatmap[x-1][y])
        if(x<len(seatmap)-1):
            neighbours.append(seatmap[x+1][y])
        if(y<len(seatmap[x])-1):
            neighbours.append(seatmap[x][y+1])
        if(y>0):
            neighbours.append(seatmap[x][y-1])
        return neighbours

    def __deepcopy__(self,memo):
        new_Seat = Seat(self.id, self.seat_letter, self.coord, self.seatclass)
        if(self.occupied):
            new_Seat.occupy(self.passenger)
        return new_Seat


def import_passengers(path="passengers.json"):
    with open(path, 'r') as f:
        json_content = json.load(f)

    passengers = []

    for pog in json_content["passengers"]:
        if isinstance(pog, list):
            passenger_group = []
            for passenger in pog:
                passenger_group.append(Passenger(passenger))
            group = PassengerGroup(passenger_group)
            passengers.append(group)
        else:
            passengers.append(Passenger(pog))
    return passengers


def import_flights(path='flights.json'):
    with open(path, 'r') as f:
        json_content = json.load(f)

    flights = []

    for flight in json_content:
        flights.append(Flight(flight))

    return flights


def split_passengers_by_destination(passengers):
    passengers_by_destination = {}
    for passenger in passengers:
        if not passenger.destination in passengers_by_destination:
            passengers_by_destination[passenger.destination] = []
        elif not isinstance(passengers_by_destination[passenger.destination], list):
            passengers_by_destination[passenger.destination] = []
        passengers_by_destination[passenger.destination].append(passenger)
    return passengers_by_destination


def split_passengers_by_time(passengers):
    passengers_by_time = {}
    for passenger in passengers:
        if not passenger.time in passengers_by_time:
            passengers_by_time[passenger.time] = []
        elif not isinstance(passengers_by_time[passenger.time], list):
            passengers_by_time[passenger.time] = []
        passengers_by_time[passenger.time].append(passenger)
    return passengers_by_time


def split_passengers_per_destination_by_time(destinations):
    passengers_by_time_destination = {}
    for destname in destinations:
        dest = destinations[destname]
        passengers_by_time_destination[destname] = split_passengers_by_time(dest)
    return passengers_by_time_destination

def sort_passengers_by_groupsize(passengers):
    for i in range(len(passengers)):
        x = passengers[i]
        j = i
        while j > 0 and passengers[j-1].size() > x.size():
            passengers[j] = passengers[j-1]
            j = j-1
        passengers[j] = x
    return passengers

def sort_grouped_passengers(passengers_by_time_destination):
    passengers_grouped_sorted = {}
    for dest in passengers_by_time_destination:
        passengers_grouped_sorted[dest] = {}
        for time in passengers_by_time_destination[dest]:
            passengers_grouped_sorted[dest][time] = sort_passengers_by_groupsize(passengers_by_time_destination[dest][time])
    return passengers_grouped_sorted


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def check_seat_pref(seat):
    if(seat.passenger.seat_pref == seat.seat_pref):
        return 1
    else:
        return 0

def check_group_alignment(seatmap, seat):
    x, y = seat.coord
    group_id = seat.passenger.group_id
    neighbours = seat.get_neigbours(seatmap)
    opt = 0
    for neigbour in neighbours:
        if(neigbour.occupied):
            if(neigbour.passenger.group_id==group_id):
                opt+=1
    if(opt>0):
        return 1

def check_seat_occupied(seat):
    if(seat.occupied):
        return 1
    else:
        return 0

def calculate_opt_count(left_passengers, seatmap):
    opt = 0
    #print(left_passengers)
    for row in seatmap:
        for seat in row:
            if(seat.occupied):
                opt += check_seat_pref(seat)
                #opt += check_group_alignment(seatmap, seat)
                opt += check_seat_occupied(seat)
                check_group_alignment(seatmap, seat)
    #print(opt)
    return opt

def max_opt(passengers):
    count = 0
    checked_facts = 2
    for passenger in passengers:
        count += checked_facts
    return count

def calculate_combinations(passengers, seatmap):
    global combinations_tried
    global rank
    global size
    global comm
    global maximum_opt
    maximum_opt = max_opt(passengers)
    #print(combinations_tried)
    #pp.pprint(passengers_old)
    results = []
    seats = []
    mpi_seats = []

    if(len(passengers)==0):
        #print("end")
        return []
    passenger = passengers[0]
    if(rank==0):
        for row in seatmap:
            if(not seatmap.next_free_seat()):
                print("no more seats")
                break
            for seat in row:
                if((not seat.occupied) and seat.seatclass == passenger.seat_class):
                    seats.append(seat.coord)
        #print("seats: "+str(len(seats)))
        if len(seats)<size:
            unused_procs = size-len(seats)
            for i in range(len(seats)):
                seat = seats.pop()
                chunc = [seat]
                mpi_seats.append(chunc)
            for i in range(unused_procs):
                chunc = []
                mpi_seats.append(chunc)
        else:
            chunc_size = int(len(seats)/size)
            #print(chunc_size)
            rest = len(seats) % size
            for i in range(size):
                chunc = []
                if(i<rest):
                    #print(i)
                    chunc.append(seats.pop())
                for j in range(chunc_size):
                    if(len(seats)>0):
                        seat = seats.pop()
                        chunc.append(seat)
                mpi_seats.append(chunc)
    else:
        seats = []
    #if(rank == 0):
        #print(size)
        #print(len(mpi_seats))
    #print("proc: "+ str(rank))
    new_seats = []
    maximum_opt = comm.bcast(maximum_opt, root=0)
    mpi_seats = comm.scatter(mpi_seats, root=0)
    results = []
    #print(len(mpi_seats))
    full_results=[]
    if(len(mpi_seats)>0):
        #print(mpi_seats)
        results = calculate_combinations_mpi(passengers,seatmap, mpi_seats)
        #print("Proc "+str(rank)+" finished combis")
        #print("len of " +str(rank)+": "+ str(len(results)))
    results = comm.gather(results, root=0)
    #print("finished gathering")
    if rank==0:
        #print(len(results))
        full_results = []
        for result in results:
            #print("len1: "+str(len(result)))
            #if type(result)==type([]):
            full_results.extend(result)
            #print("len full: "+str(len(full_results)))
        return full_results
    else:
        return []


def calculate_combinations_mpi(passengers_old, seatmap_old, mpi_seats):
    global combinations_tried
    global rank
    global size
    results = []
    seats = mpi_seats#[rank]
    if(len(passengers_old)==0):
        #print("end")
        return []
    passenger = passengers_old[0]

    #print("Seats: "+str(mpi_seats))
    #print("seat: "+str(mpi_seats[0]))
    for x,y in mpi_seats:
        seat = seatmap_old[x][y]
        seat_id = seat.id
        pass_id = passenger.id
        passengers_new = cp.copy(passengers_old)
        seatmap_new = cp.deepcopy(seatmap_old)
        seatmap_new.occupy_by_id(passenger, seat.id)
        pass_index = passengers_new.index(passenger)
        passengers_new.pop(pass_index)
        combinations_tried = combinations_tried + 1
        #opt = check_opt(seatmap_new)
        #print(seatmap_new)
        results.append([passengers_new, seatmap_new])
        result = calculate_combinations_rec(passengers_new,seatmap_new)
        if(len(result)>=2):
            if(type(result[1])==type(True)):
                if(result[1]==True):
                    #print("received best")
                    #print(result)
                    passengers, seatmap = result[0]
                    return [(passengers, seatmap)]
        results.extend(result)
    #print(result)
    return results


def calculate_combinations_rec(passengers_old, seatmap_old):
    global rank
    global maximum_opt
    if(len(passengers_old)==0):
        #print("end")
        return []
    passenger = passengers_old[0]
    #print("placing: " + str(passenger))
    #pp.pprint(passengers_old)
    done = False
    results = []
    #print("FOR by "+str(rank))
    for row in seatmap_old:
        if(not seatmap_old.next_free_seat()):
            print("no more seats")
            break
        for seat in row:
            if((not seat.occupied) and seat.seatclass == passenger.seat_class):
                seat_id = seat.id
                pass_id = passenger.id
                passengers_new = cp.copy(passengers_old)
                seatmap_new = cp.deepcopy(seatmap_old)
                seatmap_new.occupy_by_id(passenger, seat.id)
                pass_index = passengers_new.index(passenger)
                passengers_new.pop(pass_index)
                global combinations_tried
                combinations_tried = combinations_tried + 1
                #opt = check_opt(seatmap_new)
                #print(seatmap_new)
                results.append((passengers_new, seatmap_new))
                result = calculate_combinations_rec(passengers_new,seatmap_new)
                if(len(result)>=2):
                    if(type(result[1])==type(True)):
                        if(result[1]==True):
                            return result
                results.extend(result)
                #print(str(rank) + " did Nr " + str(combinations_tried))
    if(len(result)>=2):
        if(type(result[1])==type(True)):
            if(result[1]==True):
                return results
    if len(results)>0:
        opt, passengers, seatmap = calculate_best_at_rec_end(results)
        if(opt == maximum_opt):
            #print("Found best")
            return [(passengers,seatmap),True]
    return [(passengers, seatmap)]


def calculate_best_at_rec_end(maps):
    global rank
    best = (0, None)
    #print("map: "+str(maps))
    #print("maps Len: "+str(len(maps[0])))
    for cmap in maps:
        #print(maps.index(cmap))
        #print(maps)
        #print(cmap[0])
        #print(rank)
        #gprint(cmap)
        left_passengers, seatmap = cmap
        opt = calculate_opt_count(left_passengers, seatmap)
        #print(opt)
        if(opt > best[0]):
            best = opt, left_passengers, seatmap
    opt, passengers, seatmap = best
    #return passengers, seatmap
    return best



#input: array of not placed passengers, seatmap with placed passengers
def calculate_best(maps):
    global rank
    best = (0, None)
    #print("maps Len: "+str(len(maps[0])))
    for cmap in maps:
        #print(maps.index(cmap))
        #print(maps)
        #print(cmap[0])
        #print(rank)
        left_passengers, seatmap = cmap
        opt = calculate_opt_count(left_passengers, seatmap)
        #print(opt)
        if(opt > best[0]):
            best = opt, seatmap
    return best, left_passengers


def calculate_groups(passengers, flights):
    global rank
    global size
    for dest in flights:
        # check if there are passengers for this dest
        if(not dest in passengers):
            continue
        for time in flights[dest]:
            # check if there are passengers for this dest/time
            if(not time in passengers[dest]):
                continue
            #synchronize to use same dest/time
            dest_mpi = comm.bcast(dest, root=0)
            time_mpi = comm.bcast(time, root=0)
            flight = flights[dest_mpi][time_mpi][0]
            seatmap = flight.seatmap
            if rank == 0:
                print("calculate possible placements for flight to "+ dest_mpi + " int the " + time_mpi)
                #print(passengers[dest_mpi][time_mpi])
            results = calculate_combinations(passengers[dest_mpi][time_mpi], seatmap)
            #print("RESULTS " + str(dest) + " at " + str(time))
            #print(results)
            #for passengers, seatmap in results:
            #    string = "left passengers: \n" + str(passengers) + "\n" + "seatmap: \n" + str(seatmap)
            #    print(string)
            #print(results[0])
            #print(results)
            if(rank==0 and len(results)>0 and len(results[0])>0):
                (best_opt, best_map), left_passengers = calculate_best(results)
                print("Best map for flight to "+str(dest)+" at "+str(time)+" with " + str(best_opt) + " Points is:\n" + str(best_map))
                print("not placed Passengers: "+str(left_passengers))
                print("__________________________________________________")
                print("\n")




def main():
    #MPI_Init()
    global comm
    global rank
    global size
    global start_time
    global combinations_tried
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    if rank==0:
        start_time = time.time()

    passengers = import_passengers()
    flights = import_flights()
    passengers_by_destination = split_passengers_by_destination(passengers)
    passengers_by_time_destination = split_passengers_per_destination_by_time(passengers_by_destination)
    #passengers_grouped_sorted = sort_grouped_passengers(passengers_by_time_destination)
    flights_splitted = split_passengers_per_destination_by_time(split_passengers_by_destination(flights))
    result = calculate_groups(passengers_by_time_destination, flights_splitted)
    combinations_tried = comm.gather(combinations_tried, root=0)
    combs = 0
    if(rank==0):
        for count in combinations_tried:
            combs += count
        print("Combinations tried: "+str(combs))

if __name__ == "__main__":
    main()
    global rank
    if(rank==0):
        print("--- %s seconds ---" % (time.time() - start_time))
